# $project_name: A Flower / $framework_str app

## Install dependencies and project

```bash
pip install -e .
```

## Run with the Simulation Engine

In the `$project_name` directory, use `flwr run` to run a local simulation:

```bash
flwr run .
```

## Run with the Deployment Engine

> \[!NOTE\]
> An update to this example will show how to run this Flower application with the Deployment Engine and TLS certificates, or with Docker.
